IMAGE_EXTENSIONS = [
    "png",
    "jpg",
    "jpeg",
    "gif",
    "tiff",
    "bmp",
    "webp",
]

# These colors follow the tab20 colormap, however it's been rearranged
# such that brighter, more modern looking colors are at the beginning
CLASS_COLORS = [
    # Dark
    "#1f77b4",  # blue
    "#ff7f0e",  # orange
    "#bcbd22",  # lime
    "#17becf",  # cyan
    "#9467bd",  # purple
    "#8c564b",  # brown
    "#e377c2",  # pink
    "#7f7f7f",  # grey
    "#2ca02c",  # green
    "#d62728",  # red
    # Light
    "#aec7e8",  # blue
    "#ffbb78",  # orange
    "#dbdb8d",  # lime
    "#9edae5",  # cyan
    "#c5b0d5",  # purple
    "#c49c94",  # brown
    "#f7b6d2",  # pink
    "#c7c7c7",  # grey
    "#98df8a",  # green
    "#ff9896",  # red
]
